import pickle
from typing import Union

from ducks.concurrent.main import (
    ConcurrentDex,
    save as c_save,
    load as c_load,
)
from ducks.frozen.main import FrozenDex, save as f_save, load as f_load
from ducks.mutable.main import Dex, save as m_save, load as m_load


def save(box: Union[Dex, FrozenDex, ConcurrentDex], filepath: str):
    """Save a Dex, FrozenDex, or ConcurrentDex to a file."""
    if type(box) is Dex:
        m_save(box, filepath)
    if type(box) is FrozenDex:
        f_save(box, filepath)
    if type(box) is ConcurrentDex:
        c_save(box, filepath)


def load(filepath: str) -> Union[Dex, FrozenDex, ConcurrentDex]:
    """Load a Dex, FrozenDex, or ConcurrentDex from a pickle file."""
    with open(filepath, "rb") as fh:
        saved = pickle.load(fh)
        if isinstance(saved, FrozenDex):
            f_load(saved)  # mutates saved
            return saved
        elif "priority" in saved:
            return c_load(saved)
        else:
            return m_load(saved)



import pickle
from typing import Any, List, Union, Iterable, Callable, Iterator, Optional, Dict

from readerwriterlock.rwlock import RWLockRead, RWLockWrite, RWLockFair

from ducks.mutable.main import Dex
from contextlib import contextmanager


"""Lock priority options"""
READERS = "readers"
WRITERS = "writers"
FAIR = "fair"


class ConcurrentDex:

    def __init__(
        self,
        objs: Optional[Iterable[Any]] = None,
        on: Iterable[Union[str, Callable]] = None,
        priority: str = READERS,
    ):
        """Contains a Dex instance and a readerwriterlock. Wraps each Dex method in a read or write lock.

        Args:
            objs: see Dex API
            on: see Dex API
            priority: 'readers', 'writers', or 'fair'. Default 'readers'. Change this according to your usage pattern.
        """
        self.priority = priority
        self.box = Dex(objs, on)
        if priority == READERS:
            self.lock = RWLockRead()
        elif priority == WRITERS:
            self.lock = RWLockWrite()
        elif priority == FAIR:
            self.lock = RWLockFair()
        else:
            raise ValueError(f"priority must be {READERS}, {WRITERS}, or {FAIR}.")
        self._indexes = self.box._indexes  # only used during testing

    @contextmanager
    def read_lock(self):
        """Lock the ConcurrentDex for reading."""
        with self.lock.gen_rlock():
            yield

    @contextmanager
    def write_lock(self):
        """Lock the ConcurrentDex for writing.

        When doing many write operations at once, it is more efficient to do::
            with cfb.read_lock():
                for item in items:
                    cfb.box.add(item)  # calls add() on the underlying Dex.

        This performs locking only once, versus calling cfb.add() which locks for each item.
        The same pattern works for update() and remove().
        """
        with self.lock.gen_wlock():
            yield

    def get_values(self, attr: Union[str, Callable]):
        """Get a read lock and perform Dex get_values()."""
        with self.read_lock():
            return self.box.get_values(attr)

    def remove(self, obj: Any):
        """Get a write lock and perform Dex.remove()."""
        with self.write_lock():
            self.box.remove(obj)

    def add(self, obj: Any):
        """Get a write lock and perform Dex.add()."""
        with self.write_lock():
            self.box.add(obj)

    def update(self, obj: Any):
        """Get a write lock and perform Dex.update()."""
        with self.write_lock():
            self.box.update(obj)

    def __len__(self) -> int:
        """Get a read lock and get length of Dex."""
        with self.read_lock():
            return len(self.box)

    def __contains__(self, obj: Any) -> bool:
        """Get a read lock and check if the item is in the Dex."""
        with self.read_lock():
            return obj in self.box

    def __iter__(self) -> Iterator:
        """Get a read lock, make a list of the objects in the Dex, and return an iter to the list."""
        with self.read_lock():
            return iter(list(self.box))

    def __getitem__(self, query: Dict) -> List[Any]:
        """Get a read lock and perform Dex __getitem__."""
        with self.read_lock():
            return self.box[query]


def save(c_box: ConcurrentDex, filepath: str):
    """Saves a ConcurrentDex to a pickle file."""
    saved = {
        "objs": list(c_box.box.obj_map.values()),
        "on": list(c_box.box._indexes.keys()),
        "priority": c_box.priority,
    }
    with open(filepath, "wb") as fh:
        pickle.dump(saved, fh)


def load(saved: Dict) -> ConcurrentDex:
    """Creates a ConcurrentDex from the pickle file contents."""
    return ConcurrentDex(saved["objs"], saved["on"], saved["priority"])



"""
Performs object lookup for a single attribute in a FrozenDex.
"""


import numpy as np

from bisect import bisect_left, bisect_right
from typing import Union, Callable, Set

from ducks.btree import BTree
from ducks.constants import ANY, SIZE_THRESH
from ducks.utils import make_empty_array
from ducks.frozen.init_helpers import get_vals, run_length_encode


class FrozenAttrIndex:
    """
    Stores data and handles requests that are relevant to a single attribute of a FrozenDex.

    There are three places where object indexes are stored.
     - none_ids stores all indexes for with the attribute value None
     - val_to_obj_ids stores object ids for attribute values that have many objects
     - val_arr + obj_id_arr store all the rest.
    """

    def __init__(self, attr: Union[str, Callable], objs: np.ndarray, dtype: str):
        # sort the objects by attribute value, using their hashes and handling collisions
        self.dtype = dtype
        self.attr = attr

        # Nones get stored in their own special spot so they don't break sortability.
        self.none_ids = make_empty_array(self.dtype)

        # We will pull repeated attributes out into a BTree and pre-sort their indexes.
        # Saves memory, and makes object lookups *way* faster.
        self.val_to_obj_ids = BTree()

        obj_id_arr = np.arange(len(objs), dtype=self.dtype)
        for i, obj in enumerate(objs):
            obj_id_arr[i] = i
        obj_id_arr, val_arr = get_vals(objs, obj_id_arr, self.attr)

        # extract Nones. These will make the array unsortable if left in.
        none_idx = np.array(
            [i for i in range(len(val_arr)) if val_arr[i] is None], dtype=self.dtype
        )
        if len(none_idx):
            none_flag = np.zeros_like(val_arr, dtype="bool")
            none_flag[none_idx] = True
            self.none_ids = np.sort(obj_id_arr[none_flag])
            obj_id_arr = obj_id_arr[~none_flag]
            val_arr = val_arr[~none_flag]

        # Attempt to sort the values.
        sort_order = np.argsort(val_arr)  # Throws TypeError if unsortable.
        val_arr = val_arr[sort_order]
        obj_id_arr = obj_id_arr[sort_order]

        val_starts, val_run_lengths, unique_vals = run_length_encode(val_arr)
        unused = np.ones_like(obj_id_arr, dtype="bool")
        n_unused = len(unused)
        for i, val in enumerate(unique_vals):
            if val_run_lengths[i] > SIZE_THRESH:
                # extract these
                start = val_starts[i]
                end = start + val_run_lengths[i]
                unused[start:end] = False
                n_unused -= val_run_lengths[i]
                self.val_to_obj_ids[val] = np.sort(obj_id_arr[start:end])
        self.val_arr = val_arr[unused]
        self.obj_id_arr = obj_id_arr[unused]

    def get(self, val) -> np.ndarray:
        """Get indexes of objects whose attribute is val."""
        if val is ANY:
            return self.get_all()
        if val is None:
            return self.none_ids
        if val in self.val_to_obj_ids:
            return self.val_to_obj_ids[val]
        # find by bisection
        left = bisect_left(self.val_arr, val)
        if left == len(self.val_arr) or self.val_arr[left] != val:
            return make_empty_array(self.dtype)
        right = bisect_right(self.val_arr, val)
        return np.sort(self.obj_id_arr[left:right])

    def get_all(self) -> np.ndarray:
        """Get indexes of every object with this attribute. Used when matching ANY."""
        arrs = [self.obj_id_arr]
        for v in self.val_to_obj_ids.values():
            arrs.append(v)
        arrs.append(self.none_ids)
        return np.sort(np.concatenate(arrs))

    def get_values(self) -> Set:
        """Get each value we have objects for."""
        vals = set(self.val_to_obj_ids.keys())
        vals = vals.union(self.val_arr)
        if len(self.none_ids):
            vals.add(None)
        return vals

    def _get_val_arr_matches(self, lo, hi, include_lo=False, include_hi=False):
        if len(self.val_arr) == 0:
            return make_empty_array(self.dtype)

        if lo is None:
            left = 0
            lo = self.val_arr[0]
            include_lo = True
        else:
            left = bisect_left(self.val_arr, lo)

        if hi is None:
            right = len(self.val_arr)
            hi = self.val_arr[right - 1]
            include_hi = True
        else:
            right = bisect_right(self.val_arr, hi)

        # move left pointer up to fit > constraint
        if not include_lo:
            while left < len(self.val_arr) and self.val_arr[left] == lo:
                left += 1
        if left == len(self.val_arr):
            return make_empty_array(self.dtype)

        # move right pointer down to fit < constraint
        if not include_hi:
            while right > left and self.val_arr[right - 1] == hi:
                right -= 1

        small_matches = self.obj_id_arr[left:right]
        return small_matches

    def get_ids_by_range(
        self, lo, hi, include_lo=False, include_hi=False
    ) -> np.ndarray:
        """Get the object IDs associated with this value range as an Int64Set. Only usable when self.d is a tree."""
        if len(self) == 0:
            return make_empty_array(self.dtype)

        # Get matches from the val_to_obj_ids BTree
        big_matches_list = list(
            self.val_to_obj_ids.get_range(lo, hi, include_lo, include_hi)
        )

        small_matches = self._get_val_arr_matches(lo, hi, include_lo, include_hi)

        # do return
        if len(big_matches_list) == 1 and len(small_matches) == 0:
            # each big_matches is stored pre-sorted, no need to sort
            return big_matches_list[0]

        # concat all arrays and sort
        matches = np.sort(np.concatenate([small_matches] + big_matches_list))
        return matches

    def __len__(self):
        return len(self.val_arr) + len(self.val_to_obj_ids) + len(self.none_ids)



from ducks.constants import ANY
from ducks.concurrent.main import ConcurrentDex, FAIR, READERS, WRITERS
from ducks.exceptions import MissingAttribute
from ducks.frozen.main import FrozenDex
from ducks.mutable.main import Dex
from ducks.pickling import load, save



from bisect import bisect_left
from typing import Optional, Any, Dict, Union, Callable, Iterable, Set

import pickle
import numpy as np
import sortednp as snp

from ducks.btree import range_expr_to_args
from ducks.frozen.frozen_attr import FrozenAttrIndex
from ducks.frozen.utils import snp_difference
from ducks.utils import (
    make_empty_array,
    split_query,
    standardize_expr,
    validate_query,
    validate_and_standardize_operators,
)


class FrozenDex:

    def __init__(self, objs: Iterable[Any], on: Iterable[Union[str, Callable]]):
        """Create a FrozenDex containing the ``objs``, queryable by the ``on`` attributes.

        Args:
            objs: The objects that FrozenDex will contain.

            on: The attributes that will be used for finding objects.
                Must contain at least one.

        It's OK if the objects in ``objs`` are missing some or all of the attributes in ``on``.

        For the objects that do contain the attributes on ``on``, those attribute values must be hashable and sortable.
        Most Python objects are hashable. Implement the function ``__lt__(self, other)`` to make a class sortable.
        An attribute value of ``None`` is acceptable as well, even though None is not sortable.
        """
        if not on:
            raise ValueError("Need at least one attribute.")
        if isinstance(on, str):
            on = [on]

        self.obj_arr = np.empty(len(objs), dtype="O")
        self.dtype = "uint32" if len(objs) < 2 ** 32 else "uint64"
        for i, obj in enumerate(objs):
            self.obj_arr[i] = obj

        self._indexes = {}
        for attr in on:
            self._indexes[attr] = FrozenAttrIndex(attr, self.obj_arr, self.dtype)

        # only used during contains() checks
        self.sorted_obj_ids = np.sort([id(obj) for obj in self.obj_arr])

    def _find(
        self,
        match: Optional[Dict[Union[str, Callable], Any]] = None,
        exclude: Optional[Dict[Union[str, Callable], Any]] = None,
    ) -> np.ndarray:
        """Find objects in the FrozenDex that satisfy the match and exclude constraints.

        Args:
            match: Dict of ``{attribute: expression}`` defining the subset of objects that match.
                If ``None``, all objects will match.

                Each attribute is a string or Callable. Must be one of the attributes specified in the constructor.

                The expression can be any of the following:
                 - A dict of ``{operator: value}``, such as ``{'==': 1}`` ``{'>': 5}``, or ``{'in': [1, 2, 3]}``.
                 - A single value, which is a shorthand for `{'==': value}`.
                 - A list of values, which is a shorthand for ``{'in': [list_of_values]}``.
                 - ``ducks.ANY``, which matches all objects having the attribute.

                 Valid operators are '==' 'in', '<', '<=', '>', '>='.
                 The aliases 'eq' 'lt', 'le', 'lte', 'gt', 'ge', and 'gte' work too.
                 To match a None value, use ``{'==': None}``. There is no separate operator for None values.

            exclude: Dict of ``{attribute: expression}`` defining the subset of objects that do not match.
                If ``None``, no objects will be excluded.

                Each attribute is a string or Callable. Must be one of the attributes specified in the constructor.
                Valid expressions are the same as in ``match``.

        Returns:
            Numpy array of objects matching the constraints. Array will be in the same order as the original objects.
        """
        # validate input and convert expressions to dict
        validate_query(self._indexes, match, exclude)
        for arg in [match, exclude]:
            if arg:
                for key in arg:
                    arg[key] = standardize_expr(arg[key])

        # perform 'match' query
        if match:
            hit_arrays = []
            for attr, expr in match.items():
                hit_array = self._match_attr_expr(attr, expr)
                if len(hit_array) == 0:
                    # this attr had no matches, therefore the intersection will be empty. We can stop here.
                    return make_empty_array("O")
                hit_arrays.append(hit_array)

            # intersect all the hit_arrays, starting with the smallest
            for i, hit_array in enumerate(sorted(hit_arrays, key=len)):
                if i == 0:
                    hits = hit_array
                else:
                    hits = snp.intersect(hits, hit_array)
        else:
            hits = np.arange(len(self.obj_arr), dtype=self.dtype)

        # perform 'exclude' query
        if exclude:
            exc_arrays = []
            for attr, expr in exclude.items():
                exc_arrays.append(self._match_attr_expr(attr, expr))

            # subtract each of the exc_arrays, starting with the largest
            for exc_array in sorted(exc_arrays, key=len, reverse=True):
                hits = snp_difference(hits, exc_array)
                if len(hits) == 0:
                    break

        return self.obj_arr[hits]

    def _match_attr_expr(self, attr: Union[str, Callable], expr: dict) -> np.ndarray:
        """Look at an attr, handle its expr appropriately"""
        validate_and_standardize_operators(expr)
        matches = None
        # handle 'in' and '=='
        eq_expr = {op: val for op, val in expr.items() if op in ["==", "in"]}
        for op, val in eq_expr.items():
            if op == "==":
                op_matches = self._indexes[attr].get(val)
            elif op == "in":
                op_matches = self._match_any_value_in(attr, expr["in"])
            matches = (
                op_matches if matches is None else snp.intersect(op_matches, matches)
            )

        # handle range query
        range_expr = {
            op: val for op, val in expr.items() if op in ["<", ">", "<=", ">="]
        }
        if range_expr:
            min_key, max_key, include_min, include_max = range_expr_to_args(range_expr)
            range_matches = self._indexes[attr].get_ids_by_range(
                min_key, max_key, include_min, include_max
            )
            matches = (
                range_matches
                if matches is None
                else snp.intersect(range_matches, matches)
            )
        return matches

    def get_values(self, attr: Union[str, Callable]) -> Set:
        """Get the set of unique values we have for the given attribute.

        Args:
            attr: The attribute to get values for.

        Returns:
            Set of all unique values for this attribute.
        """
        return self._indexes[attr].get_values()

    def _match_any_value_in(
        self, attr: Union[str, Callable], values: Iterable[Any]
    ) -> np.ndarray:
        """"Get the union of object ID matches for the values."""
        matches = [self._indexes[attr].get(v) for v in values]
        if matches:
            return np.sort(np.concatenate(matches))
        else:
            return make_empty_array(self.dtype)

    def __contains__(self, obj):
        obj_id = id(obj)
        idx = bisect_left(self.sorted_obj_ids, obj_id)
        if idx < 0 or idx >= len(self.sorted_obj_ids):
            return False
        return self.sorted_obj_ids[idx] == obj_id

    def __iter__(self):
        return iter(self.obj_arr)

    def __len__(self):
        return len(self.obj_arr)

    def __getitem__(self, query: Dict) -> np.ndarray:
        """Find objects in the FrozenDex that satisfy the constraints.

                Args:
                    query: Dict of ``{attribute: expression}`` defining the subset of objects that match.
                        If ``{}``, all objects will match.

                        Each attribute is a string or Callable. Must be one of the attributes specified in the constructor.

                        The expression can be any of the following:
                         - A dict of ``{operator: value}``, such as ``{'==': 1}`` ``{'>': 5}``, or ``{'in': [1, 2, 3]}``.
                         - A single value, which is a shorthand for `{'==': value}`.
                         - A list of values, which is a shorthand for ``{'in': [list_of_values]}``.

                         The expression ``{'==': ducks.ANY}`` will match all objects having the attribute.
                         The expression ``{'!=': ducks.ANY}`` will match all objects without the attribute.

                         Valid operators are '==', '!=', 'in', 'not in', '<', '<=', '>', '>='.
                         The aliases 'eq', 'ne', 'lt', 'le', 'lte', 'gt', 'ge', and 'gte' work too.
                         To match a None value, use ``{'==': None}``. There is no separate operator for None values.

                Returns:
                    Numpy array of objects matching the constraints. Array will be in the same order as the original objects.
        """
        if not isinstance(query, dict):
            raise TypeError(f"Got {type(query)}; expected a dict.")
        std_query = dict()
        for attr, expr in query.items():
            std_query[attr] = standardize_expr(expr)
        match_query, exclude_query = split_query(std_query)
        return self._find(match_query, exclude_query)


def save(box: FrozenDex, filepath: str):
    """Saves this object to a pickle file."""
    with open(filepath, "wb") as fh:
        pickle.dump(box, fh)


def load(box: FrozenDex):
    """Creates a FrozenDex from the pickle file contents."""
    # If this was created by one Python process and loaded by another, the object IDs will no longer
    # correspond to the objects. Re-create the object ID array with the correct IDs.
    box.sorted_obj_ids = np.sort([id(obj) for obj in box.obj_arr])



from array import array
from typing import Callable, Union, Dict, Any, Iterable, Optional, Hashable, Set

from cykhash import Int64Set

from ducks.constants import ANY, ARR_TYPE, ARRAY_SIZE_MAX, SET_SIZE_MIN
from ducks.utils import get_attribute
from ducks.btree import BTree


class MutableAttrIndex:
    """Stores data and handles requests that are relevant to a single attribute of a Dex."""

    def __init__(
        self, attr: Union[Callable, str], objs: Optional[Iterable[Any]] = None,
    ):
        self.attr = attr
        self.none_ids = Int64Set()  # Stores object IDs for the attribute value None
        self.tree = BTree()  # Stores object IDs for all other values
        self.n_obj_ids = 0
        if objs:
            for obj in objs:
                self.add(id(obj), obj)

    def add(self, ptr: int, obj: Any):
        """Add an object if it has this attribute."""
        val, success = get_attribute(obj, self.attr)
        if not success:
            return
        self._add_val(ptr, val)
        self.n_obj_ids += 1

    def get_obj_ids(self, val: Any) -> Int64Set:
        """Get the object IDs associated with this value as an Int64Set."""
        if val is ANY:
            return self.get_all_ids()
        if val is None:
            return self.none_ids
        ids = self.tree.get(val, Int64Set())
        if type(ids) is array:
            return Int64Set(ids)
        elif type(ids) is Int64Set:
            return ids
        else:
            return Int64Set([ids])

    def remove(self, ptr: int, obj: Any):
        """Remove a single object from the index. ptr is already known to be in the Dex.
        Runs in O(1) if obj has this attr and the value of the attr hasn't changed. O(n_keys) otherwise."""
        removed = False
        val, success = get_attribute(obj, self.attr)
        if success:
            removed = self._try_remove(ptr, val)
        if not removed:
            # do O(n) search
            for val in list(self.tree.keys()):
                removed = self._try_remove(ptr, val)
                if removed:
                    break

    def get_all_ids(self) -> Int64Set:
        """Get the ID of every object that has this attribute.
        Called when matching or excluding ``{attr: hashindex.ANY}``."""
        obj_ids = Int64Set(self.none_ids)
        for key, val in self.tree.items():
            self._add_val_to_set(val, obj_ids)
        return obj_ids

    def get_values(self) -> Set:
        """Get unique values we have objects for."""
        vals = set(self.tree.keys())
        if len(self.none_ids):
            vals.add(None)
        return vals

    def get_ids_by_range(self, expr: Dict[str, Any]):
        """Get object IDs based on less than / greater than some value"""
        obj_ids = Int64Set()
        vals = self.tree.get_range_expr(expr)
        for val in vals:
            self._add_val_to_set(val, obj_ids)
        return obj_ids

    def _add_val(self, ptr, val):
        if val is None:
            self.none_ids.add(ptr)
        elif val in self.tree:
            obj_ids = self.tree[val]
            if type(obj_ids) is Int64Set:
                self.tree[val].add(ptr)
            elif type(obj_ids) is array:
                if len(obj_ids) == ARRAY_SIZE_MAX:
                    # upgrade array -> set
                    obj_ids = Int64Set(obj_ids)
                    obj_ids.add(ptr)
                    self.tree[val] = obj_ids
                else:
                    obj_ids.append(ptr)
            else:
                # obj_ids was an int, now we have two. upgrade int -> array
                self.tree[val] = array(ARR_TYPE, [obj_ids, ptr])
        else:
            # new val, add the int
            self.tree[val] = ptr

    @staticmethod
    def _add_val_to_set(val: Any, obj_ids: Int64Set):
        """We need to do this a lot"""
        if type(val) in [array, Int64Set]:
            for v in val:
                obj_ids.add(v)
        else:
            obj_ids.add(val)

    def _try_remove(self, ptr: int, val: Hashable) -> bool:
        """Try to remove the object from self.tree[val]. Return True on success, False otherwise."""
        # handle None
        if val is None and ptr in self.none_ids:
            self.none_ids.remove(ptr)
            self.n_obj_ids -= 1
            return True

        # first, check that the ptr is in here
        if val not in self.tree:
            return False
        if type(self.tree[val]) in [array, Int64Set]:
            if ptr not in self.tree[val]:
                return False
        else:
            if self.tree[val] != ptr:
                return False

        # must be in the tree
        obj_ids = self.tree[val]
        if type(self.tree[val]) in [array, Int64Set]:
            self.tree[val].remove(ptr)
            if type(obj_ids) is array:
                if len(self.tree[val]) == 1:
                    # downgrade array -> int
                    self.tree[val] = self.tree[val][0]
            else:
                if len(self.tree[val]) < SET_SIZE_MIN:
                    # downgrade set -> array
                    self.tree[val] = array(ARR_TYPE, list(self.tree[val]))
        else:
            # downgrade int -> nothing
            del self.tree[val]
        self.n_obj_ids -= 1
        return True

    def __len__(self):
        return self.n_obj_ids



import numpy as np

from typing import List, Union, Callable, Optional, Any, Dict, Tuple

from cykhash import Int64Set

from ducks.constants import ANY, VALID_OPERATORS, OPERATOR_MAP, EXCLUDE_OPERATORS
from ducks.exceptions import AttributeNotFoundError, MissingAttribute


def get_attribute(obj: Any, attr: Union[Callable, str]) -> Tuple[Any, bool]:
    """Get the object's attribute value. Return (value, success). Unsuccessful if attribute is missing."""
    if callable(attr):
        try:
            val = attr(obj)
        except MissingAttribute:
            return None, False
    elif isinstance(obj, dict):
        try:
            val = obj[attr]
        except KeyError:
            return None, False
    else:
        try:
            val = getattr(obj, attr)
        except AttributeError:
            return None, False
    return val, True


def get_attributes(cls) -> List[str]:
    """Helper function to grab the attributes of a class"""
    return list(cls.__annotations__.keys())


def split_query(query: Dict) -> Tuple[Dict, Dict]:
    """Split query into match and exclude terms"""
    match_query = dict()
    exclude_query = dict()
    for attr, expr in query.items():
        match_expr = dict()
        exclude_expr = dict()
        for op, val in expr.items():
            if op in EXCLUDE_OPERATORS:
                # invert "not in" -> "in", etc.
                exclude_expr[EXCLUDE_OPERATORS[op]] = val
            else:
                match_expr[op] = val
        if match_expr:
            match_query[attr] = match_expr
        if exclude_expr:
            exclude_query[attr] = exclude_expr
    return match_query, exclude_query


def standardize_expr(expr: Any) -> Dict:
    """Turn a find() expr into a dict of {operator: value}."""
    if isinstance(expr, dict):
        return validate_and_standardize_operators(expr)
    if isinstance(expr, list):
        return {"in": expr}
    if isinstance(expr, set) and expr is not ANY:
        raise ValueError(f"Expression {expr} is a set. Did you mean to make a dict?")
    # otherwise, it's a value
    return {"==": expr}


def validate_and_standardize_operators(expr: Dict) -> Dict:
    std_expr = {}
    for op, val in expr.items():
        if op in OPERATOR_MAP:
            std_expr[OPERATOR_MAP[op]] = val
        else:
            std_expr[op] = val
    for op in std_expr:
        if op not in VALID_OPERATORS:
            raise ValueError(
                f"Invalid operator: {op}. Operator must be one of: {VALID_OPERATORS}."
            )
    if "<" in std_expr and "<=" in std_expr:
        raise ValueError(f"Either '<' or '<=' may be used in {expr}, not both.")
    if ">" in std_expr and ">=" in std_expr:
        raise ValueError(f"Either '>' or '>=' may be used in {expr}, not both.")
    return std_expr


def validate_query(
    indexes: Dict,
    match: Optional[Dict[Union[str, Callable], Any]] = None,
    exclude: Optional[Dict[Union[str, Callable], Any]] = None,
):
    # input validation -- check that we have an index for all desired lookups
    required_indexes = set()
    if match:
        required_indexes.update(match.keys())
    if exclude:
        required_indexes.update(exclude.keys())
    missing_indexes = required_indexes.difference(indexes)
    if missing_indexes:
        raise AttributeNotFoundError(
            f"Cannot find on: {list(missing_indexes)}. Attributes must be specified on creation."
        )


def make_empty_array(dtype: str):
    """Shorthand for making a length-0 numpy array."""
    return np.empty(0, dtype=dtype)


def cyk_intersect(s1: Int64Set, s2: Int64Set) -> Int64Set:
    """Cykhash intersections are faster on small.intersect(big); handle that appropriately.
    https://github.com/realead/cykhash/issues/7"""
    return s1.intersection(s2) if len(s1) < len(s2) else s2.intersection(s1)


def cyk_union(s1: Int64Set, s2: Int64Set) -> Int64Set:
    """Cykhash unions are faster on big.union(small); handle that appropriately.
    https://github.com/realead/cykhash/issues/7"""
    return s1.union(s2) if len(s1) > len(s2) else s2.union(s1)



SIZE_THRESH = 100

ARR_TYPE = "q"  # python array type meaning "int64": https://docs.python.org/3/library/array.html
SET_SIZE_MIN = 10
ARRAY_SIZE_MAX = 20


class MatchAnything(set):
    pass


"""
ANY allows lookups like find({'attr': ANY}), which gives all objects that have an 'attr' attribute.

Why is this a set()?
We need a value that we can do "is" comparisons on, that will only be True
when it's literally this object. set() is a simple object that satisfies this property. 
"ANY is ANY" evaluates to True, but "set() is ANY" evaluates to False.
"""
ANY = MatchAnything()

VALID_OPERATORS = [
    "==",
    "eq",
    "!=",
    "ne",
    "in",
    "not in",
    "<",
    "lt",
    "<=",
    "lte",
    "le",
    ">",
    "gt",
    ">=",
    "gte",
    "ge",
    "is",
    "is not",
]

OPERATOR_MAP = {
    "eq": "==",
    "lt": "<",
    "le": "<=",  # Python style <=
    "lte": "<=",  # ElasticSearch style <=
    "gt": ">",
    "ge": ">=",  # Python style >=
    "gte": ">=",  # ElasticSearch style >=
}

EXCLUDE_OPERATORS = {"not in": "in", "!=": "=="}



class FrozenError(Exception):
    """Raised when attempting to modify a FrozenDex"""


class AttributeNotFoundError(Exception):
    """Raised when querying an attribute we don't have"""


class MissingAttribute(Exception):
    """Raise this in your attribute functions to denote that the object is missing this attribute. Finds that
    match the attribute will never return this object. Finds that exclude the attribute will."""



import numpy as np
import sortednp as snp


def snp_difference(left: np.ndarray, right: np.ndarray):
    # difference = left - indexes_in_intersection(left, right)
    _, indexes = snp.intersect(left, right, indices=True)
    indexes_to_discard = indexes[0]
    keep_these = np.ones_like(left, dtype=bool)
    keep_these[indexes_to_discard] = False
    return left[keep_these]



import numpy as np
from typing import Union, Callable
from ducks.utils import get_attribute, make_empty_array


def get_vals(objs: np.ndarray, obj_id_arr: np.ndarray, attr: Union[Callable, str]):
    """Gets vals by attribute. Returned arrays will be shorter than input if objects are missing attributes."""
    val_arr = np.empty(len(objs), dtype="O")
    success = np.empty(len(objs), dtype=bool)
    for i, obj in enumerate(objs):
        val_arr[i], success[i] = get_attribute(obj, attr)

    val_arr = val_arr[success]
    obj_id_arr = obj_id_arr[success]
    return obj_id_arr, val_arr


def run_length_encode(arr: np.ndarray):
    """
    Find counts of each element in the arr (sorted) via run-length encoding.

    Takes 10ms for 1M objs.
    """
    if len(arr) == 0:
        return (
            make_empty_array("int64"),
            make_empty_array("int64"),
            make_empty_array("int64"),
        )
    mismatch_val = arr[1:] != arr[:-1]
    change_pts = np.append(np.where(mismatch_val), len(arr) - 1)
    counts = np.diff(np.append(-1, change_pts))
    starts = np.cumsum(np.append(0, counts))[:-1]
    return starts, counts, arr[change_pts]



import pickle
from operator import itemgetter
from typing import Optional, List, Any, Dict, Callable, Union, Iterable, Set

from cykhash import Int64Set

from ducks.utils import (
    cyk_intersect,
    cyk_union,
    split_query,
    standardize_expr,
    validate_query,
)
from ducks.mutable.mutable_attr import MutableAttrIndex


class Dex:

    def __init__(
        self,
        objs: Optional[Iterable[Any]] = None,
        on: Iterable[Union[str, Callable]] = None,
    ):
        """
        Create a Dex containing the ``objs``, queryable by the ``on`` attributes.

        Args:
            objs: The objects that Dex will contain initially. Optional.

            on: The attributes that will be used for finding objects.
                Must contain at least one.

        It's OK if the objects in ``objs`` are missing some or all of the attributes in ``on``.

        For the objects that do contain the attributes in ``on``, those attribute values must be hashable and sortable.
        Most Python objects are hashable. Implement the function ``__lt__(self, other)`` to make a class sortable.
        An attribute value of ``None`` is acceptable as well, even though None is not sortable.
        """
        if not on:
            raise ValueError("Need at least one attribute.")
        if isinstance(on, str):
            on = [on]

        if objs:
            self.obj_map = {id(obj): obj for obj in objs}
        else:
            self.obj_map = dict()

        # Build an index for each attribute
        self._indexes = {}
        for attr in on:
            self._indexes[attr] = MutableAttrIndex(attr, objs)

    def _find(
        self,
        match: Dict[Union[str, Callable], Dict[str, Any]],
        exclude: Dict[Union[str, Callable], Dict[str, Any]],
    ) -> List:
        """Find objects in the Dex that satisfy the match and exclude constraints.

        Args:
            match: Dict of ``{attribute: expression}`` defining the subset of objects that match.
                If ``None``, all objects will match.

                Each attribute is a string or Callable. Must be one of the attributes specified in the constructor.

                The expression can be any of the following:
                 - A dict of ``{operator: value}``, such as ``{'==': 1}`` ``{'>': 5}``, or ``{'in': [1, 2, 3]}``.
                 - A single value, which is a shorthand for `{'==': value}`.
                 - A list of values, which is a shorthand for ``{'in': [list_of_values]}``.

                 The special value ``ducks.ANY`` will match all objects having the attribute.

                 Valid operators are '==' 'in', '<', '<=', '>', '>='.
                 The aliases 'eq' 'lt', 'le', 'lte', 'gt', 'ge', and 'gte' work too.
                 To match a None value, use ``{'==': None}``. There is no separate operator for None values.

            exclude: Dict of ``{attribute: expression}`` defining the subset of objects that do not match.
                If ``None``, no objects will be excluded.

                Each attribute is a string or Callable. Must be one of the attributes specified in the constructor.
                Valid expressions are the same as in ``match``.

        Returns:
            List of objects matching the constraints. List will be unordered.
        """
        # validate input and convert expressions to dict
        validate_query(self._indexes, match, exclude)
        obj_ids = self._find_ids(match, exclude)
        return self._obj_ids_to_objs(obj_ids)

    def add(self, obj: Any):
        """Add the object, evaluating any attributes and storing the results.
        If the object is already present, it will not be updated."""
        ptr = id(obj)
        if ptr in self.obj_map:
            return
        self.obj_map[ptr] = obj
        for attr in self._indexes:
            self._indexes[attr].add(ptr, obj)

    def remove(self, obj: Any):
        """Remove the object. Raises KeyError if not present."""
        ptr = id(obj)
        if ptr not in self.obj_map:
            raise KeyError

        for attr in self._indexes:
            self._indexes[attr].remove(ptr, obj)
        del self.obj_map[ptr]

    def update(self, obj: Any):
        """Remove and re-add the object, updating all stored attributes. Raises KeyError if object not present."""
        self.remove(obj)
        self.add(obj)

    def get_values(self, attr: Union[str, Callable]) -> Set:
        """Get the unique values we have for the given attribute.

        Args:
            attr: The attribute to get values for.

        Returns:
            Set of all unique values for this attribute.
        """
        return self._indexes[attr].get_values()

    def _find_ids(
        self,
        match: Optional[Dict[Union[str, Callable], Dict]] = None,
        exclude: Optional[Dict[Union[str, Callable], Dict]] = None,
    ) -> Int64Set:
        """Perform lookup based on given constraints. Return a set of object IDs."""
        # perform 'match' query
        if match:
            # find intersection of each attr
            hit_sets = []
            for attr, expr in match.items():
                hit_set = self._match_attr_expr(attr, expr)
                if len(hit_set) == 0:
                    # this attr had no matches, therefore the intersection will be empty. We can stop here.
                    return Int64Set()
                hit_sets.append(hit_set)

            for i, hit_set in enumerate(sorted(hit_sets, key=len)):
                # intersect this attr's hits with our hits so far
                if i == 0:
                    hits = hit_set
                else:
                    hits = cyk_intersect(hits, hit_set)
        else:
            # 'match' is unspecified, so match all objects
            hits = Int64Set(self.obj_map.keys())

        # perform 'exclude' query
        if exclude:
            exc_sets = []
            for attr, expr in exclude.items():
                exc_sets.append(self._match_attr_expr(attr, expr))

            for exc_set in sorted(exc_sets, key=len, reverse=True):
                hits = Int64Set.difference(hits, exc_set)
                if len(hits) == 0:
                    break

        return hits

    def _match_attr_expr(
        self, attr: Union[str, Callable], expr: Dict[str, Any]
    ) -> Int64Set:
        """Look at an attr, handle its expr appropriately"""
        matches = None
        # handle 'in' and '=='
        eq_expr = {op: val for op, val in expr.items() if op in ["==", "in"]}
        for op, val in eq_expr.items():
            if op == "==":
                op_matches = self._indexes[attr].get_obj_ids(val)
            elif op == "in":
                op_matches = self._match_any_value_in(attr, expr["in"])
            matches = (
                op_matches if matches is None else cyk_intersect(op_matches, matches)
            )

        # handle range query
        range_expr = {
            op: val for op, val in expr.items() if op in ["<", ">", "<=", ">="]
        }
        if range_expr:
            range_matches = self._indexes[attr].get_ids_by_range(range_expr)
            matches = (
                range_matches
                if matches is None
                else cyk_intersect(range_matches, matches)
            )
        return matches

    def _match_any_value_in(
        self, attr: Union[str, Callable], values: Iterable[Any]
    ) -> Int64Set:
        """Handle 'in' queries. Return the union of object ID matches for the values."""
        matches = Int64Set()
        for v in values:
            v_matches = self._indexes[attr].get_obj_ids(v)
            matches = cyk_union(matches, v_matches)
        return Int64Set(matches)

    def _obj_ids_to_objs(self, obj_ids: Int64Set) -> List[Any]:
        """Look up each obj_id in self.obj_map, and return the list of objs."""
        # Using itemgetter is about 10% faster than doing a comprehension like [self.objs[ptr] for ptr in hits]
        if len(obj_ids) == 0:
            return []
        elif len(obj_ids) == 1:
            return [
                itemgetter(*obj_ids)(self.obj_map)
            ]  # itemgetter returns a single item here, not in a collection
        else:
            return list(
                itemgetter(*obj_ids)(self.obj_map)
            )  # itemgetter returns a tuple of items here, so make it a list

    def __contains__(self, obj: Any):
        return id(obj) in self.obj_map

    def __iter__(self):
        return iter(self.obj_map.values())

    def __len__(self):
        return len(self.obj_map)

    def __getitem__(self, query: Dict) -> List[Any]:
        """Find objects in the Dex that satisfy the constraints.

                Args:
                    query: Dict of ``{attribute: expression}`` defining the subset of objects that match.
                        If ``{}``, all objects will match.

                        Each attribute is a string or Callable. Must be one of the attributes specified in the constructor.

                        The expression can be any of the following:
                         - A dict of ``{operator: value}``, such as ``{'==': 1}`` ``{'>': 5}``, or ``{'in': [1, 2, 3]}``.
                         - A single value, which is a shorthand for `{'==': value}`.
                         - A list of values, which is a shorthand for ``{'in': [list_of_values]}``.

                         The expression ``{'==': ducks.ANY}`` will match all objects having the attribute.
                         The expression ``{'!=': ducks.ANY}`` will match all objects without the attribute.

                         Valid operators are '==', '!=', 'in', 'not in', '<', '<=', '>', '>='.
                         The aliases 'eq', 'ne', 'lt', 'le', 'lte', 'gt', 'ge', and 'gte' work too.
                         To match a None value, use ``{'==': None}``. There is no separate operator for None values.

                Returns:
                    List of objects matching the constraints. List will be unordered.
        """
        if not isinstance(query, dict):
            raise TypeError(f"Got {type(query)}; expected a dict.")
        std_query = dict()
        for attr, expr in query.items():
            std_query[attr] = standardize_expr(expr)
        match_query, exclude_query = split_query(std_query)
        return self._find(match_query, exclude_query)


def save(box: Dex, filepath: str):
    """Saves this object to a pickle file."""
    # We can't pickle this easily, because:
    # - Int64Sets cannot be pickled, so the MutableAttrIndex is hard to save.
    # - Object IDs are specific to the process that created them, so the object map will be invalid if saved.
    # Therefore, this just pickles the objects and the list of what to build indexes on.
    # The Dex container will be built anew with __init__ on load.
    # A bit slow, but it's simple, guaranteed to work, and is very robust against changes in the container code.
    saved = {"objs": list(box.obj_map.values()), "on": list(box._indexes.keys())}
    with open(filepath, "wb") as fh:
        pickle.dump(saved, fh)


def load(saved: Dict) -> Dex:
    """Creates a Dex from the pickle file."""
    return Dex(saved["objs"], saved["on"])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ducks', 'ducks.concurrent', 'ducks.frozen', 'ducks.mutable']

package_data = \
{'': ['*']}

install_requires = \
['BTrees>=4.10.0,<5.0.0',
 'cykhash>=2.0.0,<3.0.0',
 'numpy>=1.14,<2.0',
 'readerwriterlock>=1.0.9,<2.0.0',
 'sortednp>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'ducks',
    'version': '0.5.1',
    'description': 'Provides Dex, a Python container for indexing objects of any type.',
    'long_description': "=========\nducks  🦆\n=========\n\nIndex your Python objects for fast lookup by their attributes.\n\n.. image:: https://github.com/manimino/ducks/workflows/tests/badge.svg\n    :target: https://github.com/manimino/ducks/actions\n    :alt: tests Actions Status\n.. image:: https://img.shields.io/static/v1?label=Coverage&message=100%&color=2ea44f\n    :target: https://github.com/manimino/ducks/blob/main/test/cov.txt\n    :alt: Coverage - 100%\n.. image:: https://img.shields.io/static/v1?label=license&message=MIT&color=2ea44f\n    :target: https://github.com/manimino/ducks/blob/main/LICENSE\n    :alt: license - MIT\n.. image:: https://img.shields.io/static/v1?label=python&message=3.7%2B&color=2ea44f\n    :target: https://github.com/manimino/ducks/\n    :alt: python - 3.7+\n\n-------\nInstall\n-------\n\n.. code-block::\n\n    pip install ducks\n\n-----\nUsage\n-----\n\nThe main container in ducks is called Dex.\n\n.. code-block::\n\n    from ducks import Dex\n\n    # make some objects\n    objects = [\n        {'x': 3, 'y': 'a'},\n        {'x': 6, 'y': 'b'},\n        {'x': 9, 'y': 'c'}\n    ]\n\n    # Create a Dex containing the objects.\n    # Index on x and y.\n    dex = Dex(objects, ['x', 'y'])\n\n    # match objects\n    dex[{\n        'x': {'>': 5, '<': 10},  # where 5 < x < 10\n        'y': {'in': ['a', 'b']}  # and y is 'a' or 'b'\n    }]\n    # result: [{'x': 6, 'y': 'b'}]\n\nThis is a Dex of dicts, but the objects can be any type.\n\nDex supports ==, !=, in, not in, <, <=, >, >=.\n\nThe indexes can be dict keys, object attributes, or custom functions.\n\n------------\nIs Dex fast?\n------------\n\nYes. Here's how Dex compares to other object-finders on an example task.\n\n.. image:: https://raw.githubusercontent.com/manimino/ducks/main/docs/img/perf_bench.png\n    :width: 600\n\n`Benchmark source <https://github.com/manimino/ducks/blob/main/examples/perf_demo.ipynb>`_\n\nThe closest thing to a Dex is an in-memory SQLite. While SQLite is a fantastic database, it requires\nmore overhead. As such, Dex is generally faster.\n\n----\nDocs\n----\n\nThere's more to ducks than making a Dex of dicts. `Continue in the docs. <https://ducks.readthedocs.io/en/latest/quick_start.html>`_\n",
    'author': 'Theo Walker',
    'author_email': 'theo.ca.walker@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/manimino/ducks/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from BTrees.OOBTree import OOBTree

from typing import List, Dict, Any, Tuple


class BTree:
    """
    Wraps an OOBTree instance. Tweaks it a bit:
     - BTrees len() does a full tree traversal, which is very slow. So we maintain a count instead.
     - BTrees stores None values as if they were just really really small. So "x < 1" will find the Nones.
       Here instead we disallow None entirely, make it throw TypeError.
     - Provide a nice interface for using >, >=, <, <= to get value ranges.
    """

    def __init__(self, d: Dict[Any, Any] = None):
        if d:
            if None in d:
                raise TypeError(
                    "None is not allowed in BTree because it breaks comparisons."
                )
            self.tree = OOBTree(d)
            self.length = len(d)
        else:
            self.tree = OOBTree()
            self.length = 0

    def get_range_expr(self, expr: Dict[str, Any]) -> List:
        """Get values matching a range expression like {'>': 3, '<=': 5}"""
        min_key, max_key, include_min, include_max = range_expr_to_args(expr)
        return self.get_range(min_key, max_key, include_min, include_max)

    def get_range(
        self,
        min_key=None,
        max_key=None,
        include_min: bool = True,
        include_max: bool = True,
    ) -> List:
        """
        Get values in the range of [min_key, max_key]. include_min and include_max
        determine whether values for the start and end keys will be included.

        Examples:
            Get all values: None, None, True, True
            Get 1 < key < 10: 1, 10, False, False
            Get key >= 3: 3, None, True, True
        """
        if len(self) == 0:
            return []
        excludemin = not include_min
        excludemax = not include_max
        return self.tree.values(min_key, max_key, excludemin=excludemin, excludemax=excludemax)

    def get(self, key, default=None):
        return self.tree.get(key, default)

    def keys(self):
        return self.tree.keys()

    def values(self):
        return self.tree.values()

    def items(self):
        return self.tree.items()

    def __len__(self):
        return self.length

    def __setitem__(self, key, value):
        if key is None:
            raise TypeError(
                "None is not allowed in BTree because it breaks comparisons."
            )
        if len(self) == 0:
            # OOBTree oddity: it allows a non-comparable object on the first insert, but
            # if it gets one, all future inserts will fail.
            # So let's raise a TypeError if the very first insert is a non-comparable type.
            key > key
        if key not in self.tree:
            self.length += 1
        self.tree[key] = value

    def __getitem__(self, key):
        return self.tree[key]

    def __delitem__(self, key):
        self.length -= 1
        del self.tree[key]

    def __contains__(self, item):
        return item in self.tree


def range_expr_to_args(expr: Dict[str, Any]) -> Tuple[Any, Any, bool, bool]:
    """
    Turn a range expr into (min_key, max_key, include_min, include_max), which are easier to use with BTrees.
    e.g., translates {'<': 3} into get_values(3, None, True, False).
    Will ignore keys in expr other than '<', '<=', '>', '>='.
    """
    min_key = None
    max_key = None
    include_min = True
    include_max = True
    if ">" in expr:
        min_key = expr[">"]
        include_min = False
    if ">=" in expr:
        min_key = expr[">="]
        include_min = True
    if "<" in expr:
        max_key = expr["<"]
        include_max = False
    if "<=" in expr:
        max_key = expr["<="]
        include_max = True
    return min_key, max_key, include_min, include_max
